﻿using MeetingBooster.Api;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading.Tasks;
using System.Web;

namespace ApiExample {
	/*
	 * This example show how to connect to MeetingBooster using the Api.
	 * The example will get a list of mettings changed since last request and 
	 * download the metting minutes to a local folder.
	 */ 
	class Program {

		static string Url;
		static JSOAuthResponse Oauth;
		static WebClient Wc;

		/// <summary>
		/// Do the oauth authorization
		/// </summary>
		/// <param name="username">The user email. (If MeetingBooster is using AD you can specify the AD username)</param>
		/// <param name="password">The user password</param>
		/// <returns>True if authorization succeeds</returns>
		private static bool Authorize( string username,string password) {

			string args = String.Format("grant_type=password&username={0}&password={1}&app_id={2}",
				HttpUtility.UrlEncode(username, Encoding.UTF8),
				HttpUtility.UrlEncode(password, Encoding.UTF8),
				"Api");

			try {
					string response = Wc.UploadString(Url + "/oauth", args);
					Oauth = JsonConvert.DeserializeObject<JSOAuthResponse>(response);
					Wc.Headers.Add("Authorization", "Bearer " + Oauth.access_token);
			}
			catch (WebException e) {
				HttpWebResponse response = e.Response as HttpWebResponse;
				using (Stream resStream = response.GetResponseStream()) {
					StreamReader reader = new StreamReader(resStream);
					string responsedata = reader.ReadToEnd();
					JSOAuthError oautherror = JsonConvert.DeserializeObject<JSOAuthError>(responsedata);
					return false;
				}
			}
			return true;
		}

		/// <summary>
		/// Will wrap the function in a try catch and do a token refesh if needed
		/// </summary>
		/// <param name="fn">The function to wrap</param>
		private static void OAuthSafe(Action fn) {

			try {
				fn();
			}
			catch (WebException e) {
				if (!HandleTokenRefresh(e)) {
					Console.WriteLine("Error from: " + e.Response.ResponseUri);
					throw e;
				}
				fn();
			}
		}

		/// <summary>
		/// If we have a 401 Unauthorized status code - we will do an oauth token refresh request
		/// </summary>
		/// <param name="e">The WebException</param>
		/// <returns>True the token refresh succeeded</returns>
		private static bool HandleTokenRefresh(WebException e) {

			HttpWebResponse webresponse = e.Response as HttpWebResponse;
			if (webresponse == null) return false;
			if (webresponse.StatusCode != HttpStatusCode.Unauthorized) return false;

			string args = String.Format("grant_type=refresh_token&refresh_token={0}", Oauth.refresh_token);
			Wc.Headers.Remove("Authorization");
			string response = Wc.UploadString(Url + "/oauth", args);
			Oauth = JsonConvert.DeserializeObject<JSOAuthResponse>(response);
			Wc.Headers.Add("Authorization", "Bearer " + Oauth.access_token);
			return true;
		}

		/// <summary>
		/// Create a filename from a meeting.
		/// We make it from the host name, meeting start date, and meeting title
		/// </summary>
		/// <param name="meeting">The meeting</param>
		/// <returns>The filename</returns>
		private static string GetFileName(JSMeeting meeting) {

			string name = meeting.Host.Name + " " + meeting.Start.ToShortDateString() + meeting.Title;
			// Trucate the name if it is long
			if (name.Length > 120) name = name.Substring(0, 120);
			name += ".pdf";
			// Make sure the filename does not contain invalid characters
			var invalidchars = Path.GetInvalidFileNameChars();
			foreach (char ch in invalidchars) name = name.Replace(ch, '_');
			return name;
		}

		/// <summary>
		/// Get all the meetings modified since "lastmodified" and save the minutes 
		/// for each meeting in the destfolder specfied.
		/// </summary>
		/// <param name="destfolder">The folder where the minutes are saved</param>
		/// <param name="lastmodified">Is set ony meeting modified later than lastmodified will be processed</param>
		private static void GetMinutes( string destfolder,DateTime? lastmodified) {

			int skip = 0;
			do {
				string url = Url + "/api/v1/meetings/list?maxResults=5&skip=" + skip;
				if (lastmodified.HasValue)
					url += "&lastmodified=" + lastmodified.Value.ToIso8601();
				string data = "";
				OAuthSafe(() => {
					data = Encoding.UTF8.GetString( Wc.DownloadData(url));
				});
				string next = Wc.ResponseHeaders.Get("paging-next");
				var meetings = JsonConvert.DeserializeObject<JSMeeting[]>(data);
				foreach (var meeting in meetings) {
					// We will skip meeting that has not been executed
					if (!meeting.ActualEnd.HasValue) continue;
					Console.WriteLine(meeting.Title);

					string path = destfolder + "\\" + GetFileName( meeting);
					var timeZone = TimeZone.CurrentTimeZone.StandardName;
					// if the client is a different language than the server we need to hardcode the timezone:
					// timeZone = "Romance Standard Time";
					url = Url + String.Format( "/api/v1/meetings/minutes?id={0}&timezone={1}",meeting.Id,timeZone);
					OAuthSafe(() => {
						Wc.DownloadFile(url, path);
					});
				}
				if (String.IsNullOrEmpty(next)) break;
				skip = int.Parse(next);
			} while (true);
		}
		/// <summary>
		/// Save a task. The assignee id should match an existing user. If you update an existing task include the Id. If you create a new task, omit the Id.
		/// </summary>
		private static void SaveTask() {
			string url = Url + "/api/v1/tasks/save";
			var task = new {
				Id = "0DA39273-72F2-496A-832F-39D628B3D9FC",
				Assignee = new { id = "128A5AB8-F020-4269-AE3A-43123A6B0A48" },
				DueDate = "2021-05-17T08:21:10Z",
				Subject = "Remember to buy beer for birthday party",
				CompletionPercentage = 0,
				Priority = 1,
				Description = ""
			};
			var dataString = JsonConvert.SerializeObject(task);
			Wc.Headers.Add(HttpRequestHeader.ContentType, "application/json");
			OAuthSafe(() => {
				var data = Wc.UploadString(new Uri(url), "POST", dataString);
				Console.WriteLine(data);
			});
		}

		static int Main(string[] args) {

			if (args.Count() < 4) {
				Console.WriteLine("Parameters: url folder email password");
				Console.WriteLine(@"Example: http://company.meetingbooster.com c:\minutes myemail@company.com mysecret");
				Console.WriteLine("Note: The user must have read access to all the meetingss");
				return -1;
			}
			Url = args[0];
			string folder = args[1];
			string username = args[2];
			string password = args[3];
			if (!Directory.Exists(folder)) {
				Console.WriteLine( String.Format( "The destination folder \"{0}\"does not exist",folder));
				return -1;
			}
			try {
				DateTime? lastmodified = null;
				DateTime utcnow = DateTime.UtcNow;
				//If we have a settings file read the last modified from the file
				string settingspath = folder + "\\settings.txt";
				if ( File.Exists(settingspath)) {
					string settings = File.ReadAllText(settingspath);
					lastmodified = new DateTime(long.Parse(settings));
				}
				using (Wc = new WebClient()) {
					if (!Authorize(username, password)) {
						Console.Error.WriteLine("Failed to authenticate");
						return -1;
					}
					GetMinutes( folder,lastmodified);
				}
				// Save the lastmodified to the settings file
				File.WriteAllText( settingspath, utcnow.Ticks.ToString());
			}
			catch (Exception e) {
				WebException we = e as WebException;
				if (we != null && we.Response != null) {
					var response = we.Response as HttpWebResponse;
					if (response != null) {
						if (response.StatusCode == HttpStatusCode.Forbidden) {
							Console.Error.WriteLine("You do not have access. (The user must have acces to read all meetings in MeetingBooster");
							return -2;
						}
					}
				}
				Console.Error.WriteLine("Exception: " + e.Message);
				Console.Error.WriteLine(e.StackTrace);
			}
			return 0;
		}
	}
}
